package Sys::Virt::IO::Async;

use Sys::Virt::Event::IO::Async;

our $VERSION = '0.0.3';

1;

__END__

=head1 NAME

Sys::Virt::IO::Async - Helpers to integrate Sys::Virt with IO::Async

=head1 DESCRIPTION

=head1 AUTHORS

=over 4

=item * Erik Huelsmann C<< ehuels@gmail.com >>

=back

=head1 LICENSE AND COPYRIGHT

This software is copyright (c) 2023 by Erik Huelsmann.

This is free software; you can redistribute it and/or modify
it under the same terms as the Perl 5 programming language system itself.
